use anyhow::{anyhow, bail};
use i2p::net::{I2pAddr, I2pSocketAddr, ToI2pSocketAddrs};
use serde::{Deserialize, Serialize};

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct PeerId {
    i2p_addr: I2pSocketAddr,
}

impl PeerId {
    pub fn addr(&self) -> I2pSocketAddr {
        self.i2p_addr.to_owned()
    }
    pub fn addr_ref(&self) -> &I2pSocketAddr {
        &self.i2p_addr
    }
    pub fn b32_addr(&self) -> anyhow::Result<I2pAddr> {
        I2pAddr::from_b64(&self.i2p_addr.dest().string()).map_err(|e| anyhow!(e))
    }
}

impl ToString for PeerId {
    fn to_string(&self) -> String {
        self.i2p_addr.to_string()
    }
}

impl TryFrom<&str> for PeerId {
    type Error = anyhow::Error;

    fn try_from(value: &str) -> Result<Self, anyhow::Error> {
        match ToI2pSocketAddrs::to_socket_addrs(&value) {
            Ok(addr_iter) => {
                for addr in addr_iter {
                    return Ok(PeerId { i2p_addr: addr });
                }
                return Err(anyhow::Error::msg("No valid I2P address found"));
            }
            Err(e) => bail!(e),
        }
    }
}

impl TryFrom<String> for PeerId {
    type Error = anyhow::Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        Self::try_from(value.as_str())
    }
}

impl From<I2pSocketAddr> for PeerId {
    fn from(value: I2pSocketAddr) -> Self {
        PeerId { i2p_addr: value }
    }
}

impl From<PeerId> for I2pSocketAddr {
    fn from(value: PeerId) -> Self {
        value.i2p_addr
    }
}

impl Default for PeerId {
    fn default() -> Self {
        PeerId {
            i2p_addr: I2pSocketAddr::new(I2pAddr::new(""), 0),
        }
    }
}
