use std::collections::HashSet;

use anyhow::{anyhow, Error};
use bonsaidb::core::{
    connection::Connection,
    document::Emit,
    schema::{view::map::Mappings, Collection, MapReduce, SerializedCollection, View, ViewSchema},
};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{Family, FamilyId, PeerId};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "peer-families", views = [DbPeerFamilyIdByMemberId, DbPeerFamilyByMemberId])]
pub(super) struct DbPeerFamily {
    #[natural_id]
    pub(super) id: AsKey<FamilyId>,
    pub(super) name: Option<String>,
    pub(super) members: HashSet<PeerId>,
}

#[derive(Debug, Clone, View, ViewSchema)]
#[view(collection = DbPeerFamily, key = AsKey<PeerId>, value = Option<FamilyId>, name = "id-by-member-id")]
pub(super) struct DbPeerFamilyIdByMemberId;

impl MapReduce for DbPeerFamilyIdByMemberId {
    fn map<'doc>(
        &self,
        document: &'doc bonsaidb::core::document::BorrowedDocument<'_>,
    ) -> bonsaidb::core::schema::ViewMapResult<'doc, Self> {
        let content = DbPeerFamily::document_contents(document)?;
        Ok(content
            .members
            .iter()
            .fold(Mappings::default(), |acc, member| {
                acc.and(
                    document
                        .header
                        .emit_key_and_value(
                            AsKey::new(member.to_owned()),
                            Some((*content.id).clone()),
                        )
                        .unwrap_or(Mappings::none()),
                )
            }))
    }
}

#[derive(Debug, Clone, View, ViewSchema)]
#[view(collection = DbPeerFamily, key = AsKey<PeerId>, value = Option<Family>, name = "by-member-id")]
pub(super) struct DbPeerFamilyByMemberId;

impl MapReduce for DbPeerFamilyByMemberId {
    fn map<'doc>(
        &self,
        document: &'doc bonsaidb::core::document::BorrowedDocument<'_>,
    ) -> bonsaidb::core::schema::ViewMapResult<'doc, Self> {
        let content = DbPeerFamily::document_contents(document)?;
        Ok(content
            .members
            .iter()
            .fold(Mappings::default(), |acc, member| {
                acc.and(
                    document
                        .header
                        .emit_key_and_value(
                            AsKey::new(member.to_owned()),
                            Some(content.clone().into()),
                        )
                        .unwrap_or(Mappings::none()),
                )
            }))
    }
}

impl From<DbPeerFamily> for Family {
    fn from(value: DbPeerFamily) -> Self {
        Family {
            id: (*value.id).clone(),
            name: value.name,
            members: value.members.iter().map(|p| p.clone()).collect_vec(),
        }
    }
}

impl StateDB {
    pub fn add_peer_family(
        &self,
        id: FamilyId,
        name: Option<String>,
        initial_members: Vec<PeerId>,
    ) -> anyhow::Result<()> {
        DbPeerFamily::push(
            DbPeerFamily {
                id: AsKey::new(id),
                name,
                members: HashSet::from_iter(initial_members),
            },
            &self.db,
        )
        .map(|_| ())
        .map_err(|e| anyhow!(e))
    }

    pub fn add_peer_to_family(&self, peer: PeerId, family: FamilyId) -> anyhow::Result<()> {
        DbPeerFamily::get(&AsKey::new(family), &self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find peer family"))?
            .modify(&self.db, |doc| {
                doc.contents.members.insert(peer.clone());
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn set_peer_family_name(
        &self,
        family: FamilyId,
        name: Option<String>,
    ) -> anyhow::Result<()> {
        DbPeerFamily::get(&AsKey::new(family), &self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find peer family"))?
            .modify(&self.db, |doc| {
                doc.contents.name = name.clone();
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn get_peer_family(&self, id: FamilyId) -> anyhow::Result<Option<Family>> {
        DbPeerFamily::get(&AsKey::new(id), &self.db)
            .map(|doc_opt| doc_opt.map(|doc| doc.contents.into()))
            .map_err(|e| anyhow!(e))
    }

    pub fn get_family_of_peer(&self, peer: PeerId) -> anyhow::Result<Option<FamilyId>> {
        self.db
            .view::<DbPeerFamilyIdByMemberId>()
            .with_key(&AsKey::new(peer))
            .query_with_collection_docs()
            .map(|results| match results.mappings.len() {
                0 => Ok(None),
                1 => Ok(results.mappings.first().unwrap().value.clone()),
                _ => Err(Error::msg("Peer appears to be member of multiple families")),
            })?
            .map_err(|e| anyhow!(e))
    }

    pub fn get_peer_family_members(&self, peer: PeerId) -> anyhow::Result<Vec<PeerId>> {
        self.db
            .view::<DbPeerFamilyByMemberId>()
            .with_key(&AsKey::new(peer))
            .query_with_collection_docs()
            .map(|results| match results.mappings.len() {
                0 => Ok(vec![]),
                1 => Ok(results
                    .mappings
                    .first()
                    .unwrap()
                    .value
                    .clone()
                    .map(|family| family.members)
                    .unwrap_or(vec![])),
                _ => Err(Error::msg("Peer appears to be member of multiple families")),
            })?
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{Family, FamilyId, PeerId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let family_id = FamilyId::new();
        let peer_id = PeerId::default();

        db.add_peer_family(
            family_id.clone(),
            Some("My family name".to_string()),
            vec![peer_id.clone()],
        )
        .unwrap();
        let retrieved_family = db.get_peer_family(family_id.clone()).unwrap();

        assert_eq!(
            retrieved_family,
            Some(Family::new(
                family_id,
                Some("My family name".to_string()),
                vec![peer_id]
            ))
        )
    }

    #[test]
    fn set_name() {
        let db = StateDB::init(None);
        let family_id = FamilyId::new();
        let peer_id = PeerId::default();

        db.add_peer_family(
            family_id.clone(),
            Some("My family name".to_string()),
            vec![peer_id.clone()],
        )
        .unwrap();

        assert_eq!(
            db.get_peer_family(family_id.clone()).unwrap(),
            Some(Family::new(
                family_id.clone(),
                Some("My family name".to_string()),
                vec![peer_id.clone()]
            ))
        );

        db.set_peer_family_name(family_id.clone(), Some("New family name".to_string()))
            .unwrap();

        assert_eq!(
            db.get_peer_family(family_id.clone()).unwrap(),
            Some(Family::new(
                family_id,
                Some("New family name".to_string()),
                vec![peer_id]
            ))
        );
    }

    #[test]
    fn get_family_of_peer() {
        let db = StateDB::init(None);
        let family_id = FamilyId::new();
        let peer_id = PeerId::default();

        db.add_peer_family(
            family_id.clone(),
            Some("My family name".to_string()),
            vec![peer_id.clone()],
        )
        .unwrap();

        assert_eq!(db.get_family_of_peer(peer_id).unwrap(), Some(family_id))
    }
}
