use crate::{api::v0::app::App, comm::CommHandle, node_events::UbisyncNodeEvent, Ubisync};
use std::{
    collections::HashMap,
    sync::{
        mpsc::{channel, Receiver, Sender},
        Arc, Mutex, RwLock,
    },
};
use ubisync_lib::{
    api::events::AppEvent,
    messages::{Message, MessageContent},
    peer::Peer,
    types::{AppId, Element, ElementContent, ElementId, FamilyId, PeerId, PotId, Tag},
};

use anyhow::Error;
use tracing::{debug, warn};

mod api_state;
mod comm_state;
mod database;

pub use api_state::ApiState;
pub use comm_state::CommState;

use self::database::StateDB;

pub struct State {
    db: StateDB,
    comm_handle: RwLock<Option<Arc<CommHandle>>>,
    app_event_channels: RwLock<HashMap<AppId, (Sender<AppEvent>, Arc<Mutex<Receiver<AppEvent>>>)>>,
    node_event_callback: RwLock<Option<Box<dyn Fn(UbisyncNodeEvent) + Send + Sync>>>,
}

impl State {
    pub async fn new(db_path: Option<String>) -> anyhow::Result<Arc<State>> {
        let db = StateDB::init(db_path);

        let state = Arc::new(State {
            db,
            comm_handle: RwLock::new(None),
            app_event_channels: Default::default(),
            node_event_callback: RwLock::new(None),
        });
        state.init_event_channels();
        Ok(state)
    }

    pub fn set_comm_handle(&self, handle: Arc<CommHandle>) {
        *self
            .comm_handle
            .write()
            .as_deref_mut()
            .expect("Could not set state's CommHandle") = Some(handle);
    }

    pub fn set_node_event_callback<T>(&self, cb: T, ubi: Arc<Ubisync>)
    where
        T: Fn(UbisyncNodeEvent, Arc<Ubisync>) + Send + Sync + 'static,
    {
        match self.node_event_callback.write() {
            Ok(mut writeguard) => {
                *writeguard = Some(Box::new(move |ev: UbisyncNodeEvent| (cb)(ev, ubi.clone())))
            }
            Err(e) => warn!("Failed to set node event callback: {:?}", e),
        }
    }

    pub fn own_peer_id(&self) -> Option<PeerId> {
        if let Ok(guard) = self.comm_handle.read() {
            guard.as_ref().map(|t| t.own_peer_id().ok()).flatten()
        } else {
            None
        }
    }

    pub fn get_apps(&self) -> anyhow::Result<Vec<App>> {
        self.db.get_all_apps()
    }

    pub fn set_element_content(
        &self,
        element_id: ElementId,
        content: ElementContent,
    ) -> anyhow::Result<()> {
        self.db
            .set_element_content(element_id.clone(), content.clone())
            .inspect(|_| {
                //TODO: Get all peers interested in the element, e.g. because they subscribe to the element's pot, a share, etc.
                self.send_to_peers(
                    MessageContent::UpdateElement {
                        id: element_id.clone(),
                        content: content.clone(),
                    },
                    self.own_peer_id()
                        .map(|id| self.db.get_peer_family_members(id).unwrap_or_default())
                        .unwrap_or_default(),
                )
            })
    }

    pub fn remove_element(&self, element_id: ElementId) -> anyhow::Result<()> {
        self.db.remove_element(element_id.clone()).inspect(|_| {
            //TODO: Get all peers interested in the element, e.g. because they subscribe to the element's pot, a share, etc.
            self.send_to_peers(
                MessageContent::RemoveElement {
                    id: element_id.clone(),
                },
                self.own_peer_id()
                    .map(|id| self.db.get_peer_family_members(id).unwrap_or_default())
                    .unwrap_or_default(),
            )
        })
    }

    pub fn get_element(&self, id: ElementId) -> Option<Element> {
        self.db.get_element(id)
    }

    pub fn get_elements_by_tag(&self, _tag: &Tag) -> Vec<ElementId> {
        todo!()
    }

    pub fn add_peer(&self, peer: Peer) -> anyhow::Result<()> {
        self.db.add_peer(peer)
    }

    pub fn get_peers(&self) -> anyhow::Result<Vec<Peer>> {
        self.db.get_all_peers()
    }

    pub fn add_pot_member(&self, pot: PotId, app: AppId) -> anyhow::Result<()> {
        let p = self
            .db
            .get_pot(pot.clone())?
            .ok_or(Error::msg("Could not find pot"))?;
        self.db
            .add_pot_membership(pot.clone(), app.clone())
            .inspect(|_| {
                self.emit_app_event(
                    &app,
                    AppEvent::NewPot {
                        id: pot.clone(),
                        app_type: p.app_type,
                    },
                )
            })
    }

    pub fn add_family_member(&self, peer: PeerId) -> anyhow::Result<()> {
        let my_id = self
            .comm_handle
            .read()
            .map_err(|_| Error::msg("Failed to lock on CommHandle"))?
            .to_owned()
            .ok_or(Error::msg("CommHandle not initialized"))?
            .own_peer_id()?;

        if self.db.get_family_of_peer(my_id.clone())?.is_none() {
            self.db
                .add_peer_family(FamilyId::new(), None, vec![my_id.clone()])?;
        }

        self.db.add_peer_to_family(
            peer,
            self.db
                .get_family_of_peer(my_id)?
                .ok_or(Error::msg("Could not find own family"))?,
        )
    }

    pub fn get_event_receiver(
        &self,
        app: &AppId,
    ) -> anyhow::Result<Arc<Mutex<Receiver<AppEvent>>>> {
        self.create_event_channel_if_not_exists(app.clone());
        match self.app_event_channels.read() {
            Ok(map) => match map.get(&app) {
                Some((_, rx)) => Ok(rx.clone()),
                None => Err(Error::msg("Failed to get event receiver")),
            },
            Err(_) => Err(Error::msg("Failed to lock on event_channels")),
        }
    }

    pub fn emit_node_event(&self, ev: UbisyncNodeEvent) -> anyhow::Result<()> {
        debug!("1");
        match self.node_event_callback.read() {
            Ok(readguard) => {
                debug!("2");
                debug!("{:?}", readguard.is_some());
                // If a callback is set, call it.
                if let Some(cb) = readguard.as_ref() {
                    debug!("3");
                    (*cb)(ev);
                }
                // Whether a callback is set or not, return Ok(_)
                Ok(())
            }
            Err(_) => Err(Error::msg("Failed to lock on node event callback function")),
        }
    }

    fn emit_app_event(&self, app: &AppId, ev: AppEvent) {
        if let Ok(channels) = self.app_event_channels.read() {
            if let Some((sender, _)) = channels.get(app) {
                debug!("Emitting app event to {:?}", &app);
                if let Err(e) = sender.send(ev) {
                    debug!("Emitting app event failed: {:?}", e);
                }
            }
        } else {
            debug!("Failed to get channels");
        }
    }

    fn create_event_channel_if_not_exists(&self, app: AppId) {
        if let Ok(mut map) = self.app_event_channels.write() {
            if !map.contains_key(&app) {
                debug!(
                    "No event channel for {:?} exists, creating a new one.",
                    &app
                );
                let (tx, rx) = channel();
                map.insert(app, (tx, Arc::new(Mutex::new(rx))));
            }
        }
    }

    fn init_event_channels(&self) {
        let app_ids = self.db.get_all_app_ids().unwrap();
        for app in app_ids {
            self.create_event_channel_if_not_exists(app);
        }
    }

    fn send_to_peers(&self, ct: MessageContent, peers: Vec<PeerId>) {
        match self.comm_handle.read() {
            Ok(opt) => {
                if opt.is_some() {
                    let arc = opt.as_ref().unwrap().clone();
                    tokio::spawn(async move {
                        for peer in peers {
                            if let Err(e) =
                                arc.send(peer.addr_ref(), Message::new(ct.clone())).await
                            {
                                debug!(
                                    "Sending to peer '{:?}' returned an error: {}",
                                    peer.b32_addr(),
                                    e
                                )
                            }
                        }
                    });
                }
            }
            Err(e) => debug!("{}", e),
        }
    }
}
